@@ -49,6 +49,11 @@ int argcount = 0, argsize = 0;
 char **args = 0;
 
 char *custom_ere;
+regex_t hierre, tradre, excsre, classicalre, customre;
+
+static char* regex_get_error(int errcode, regex_t *compiled);
+static void  regex_compile_pattern(void);
+static void  regex_clean(void);
 
 void error(char *format, ...)
 {
@@ -133,50 +138,37 @@ void add_argument(char *newarg)
 /* True or false? Is this a valid filename? */
 int valid_name(const struct dirent *d)
 {
-  char *c = d->d_name;
-  regex_t hierre, tradre, excsre, classicalre, customre;
+    char         *s;
+    unsigned int  retval;
 
-  /* The regcomps should be moved to program init */
-  if (regex_mode == RUNPARTS_ERE) {
-    if (regcomp(&customre, custom_ere, REG_EXTENDED | REG_NOSUB)) {
-      error("custom regex failure");
-      exit(1);
-    }
-    return !regexec(&customre, c, 0, NULL, 0);
-  }
-  else if (regex_mode == RUNPARTS_LSBSYSINIT) {
+    s = (char *)&(d->d_name);
 
-    if (regcomp
-	(&hierre, "^_?([a-z0-9_.]+-)+[a-z0-9]+$", REG_EXTENDED | REG_NOSUB)
-	|| regcomp(&excsre, "^[a-z0-9-].*dpkg-(old|dist|new|tmp)$",
-		   REG_EXTENDED | REG_NOSUB)
-	|| regcomp(&tradre, "^[a-z0-9][a-z0-9-]*$", REG_NOSUB)) {
-      error("regex failure");
-      exit(1);
-    }
+    if (regex_mode == RUNPARTS_ERE)
+        retval = !regexec(&customre, s, 0, NULL, 0);
 
-    if (!regexec(&hierre, c, 0, NULL, 0))
-      return regexec(&excsre, c, 0, NULL, 0);
+    else if (regex_mode == RUNPARTS_LSBSYSINIT) {
 
-    return !regexec(&tradre, c, 0, NULL, 0);
-  }
-  else {
-    if (regcomp(&classicalre, "^[a-zA-Z0-9_-]+$", REG_EXTENDED | REG_NOSUB)) {
-      error("regex failure");
-      exit(1);
-    }
-    return !regexec(&classicalre, c, 0, NULL, 0);
-  }
+        if (!regexec(&hierre, s, 0, NULL, 0))
+            retval = regexec(&excsre, s, 0, NULL, 0);
 
+	else
+            retval = !regexec(&tradre, s, 0, NULL, 0);
+
+    } else
+        retval = !regexec(&classicalre, s, 0, NULL, 0);
+
+    return retval;
 }
 
 /* Execute a file */
 void run_part(char *progname)
 {
-  int result;
-  int pid;
+  int result, waited;
+  int pid, r;
   int pout[2], perr[2];
 
+  waited = 0;
+
   if (report_mode && (pipe(pout) || pipe(perr))) {
     error("pipe: %s", strerror(errno));
     exit(1);
@@ -207,30 +199,56 @@ void run_part(char *progname)
 
   if (report_mode) {
     fd_set set;
-    int max, r, printflag;
+    sigset_t tempmask;
+    struct timespec zero_timeout;
+    struct timespec *the_timeout;
+    int max, printflag;
     ssize_t c;
     char buf[4096];
 
+    sigemptyset(&tempmask);
+    sigprocmask(0, NULL, &tempmask);
+    sigdelset(&tempmask, SIGCHLD);
+
+    memset(&zero_timeout, 0, sizeof(zero_timeout));
+    the_timeout = NULL;
+
     close(pout[1]);
     close(perr[1]);
     max = pout[0] > perr[0] ? pout[0] + 1 : perr[0] + 1;
     printflag = 0;
 
     while (pout[0] >= 0 || perr[0] >= 0) {
+      if (!waited) {
+        r = waitpid(pid, &result, WNOHANG);
+        if (r == -1) {
+          error("waitpid: %s", strerror(errno));
+          exit(1);
+        }
+        if (r != 0 && (WIFEXITED(result) || WIFSIGNALED(result))) {
+          /* If the process dies, set a zero timeout. Rarely, some processes
+           * leak file descriptors (e.g., by starting a naughty daemon).
+           * select() would wait forever since the pipes wouldn't close.
+           * We loop, with a zero timeout, until there's no data left, then
+           * give up. This shouldn't affect non-leaky processes. */
+          waited = 1;
+          the_timeout = &zero_timeout;
+        }
+      }
 
-      do {
-	FD_ZERO(&set);
-	if (pout[0] >= 0)
-	  FD_SET(pout[0], &set);
-	if (perr[0] >= 0)
-	  FD_SET(perr[0], &set);
-	r = select(max, &set, 0, 0, 0);
-      } while (r < 0 && errno == EINTR);
+      FD_ZERO(&set);
+      if (pout[0] >= 0)
+        FD_SET(pout[0], &set);
+      if (perr[0] >= 0)
+        FD_SET(perr[0], &set);
+      r = pselect(max, &set, 0, 0, the_timeout, &tempmask);
 
       if (r < 0) {
-	/* assert(errno != EINTR) */
-	error("select: %s", strerror(errno));
-	exit(1);
+        if (errno == EINTR)
+            continue;
+
+        error("select: %s", strerror(errno));
+        exit(1);
       }
       else if (r > 0) {
 	if (pout[0] >= 0 && FD_ISSET(pout[0], &set)) {
@@ -274,6 +292,13 @@ void run_part(char *progname)
 	  }
 	}
       }
+      else if (r == 0 && waited) {
+        /* Zero timeout, no data left. */
+        close(perr[0]);
+        perr[0] = -1;
+        close(pout[0]);
+        pout[0] = -1;
+      }
       else {
 	/* assert(FALSE): select was called with infinite timeout, so
 	   it either returns successfully or is interrupted */
@@ -281,7 +306,14 @@ void run_part(char *progname)
     }				/*while */
   }
 
-  waitpid(pid, &result, 0);
+  if (!waited) {
+    r = waitpid(pid, &result, 0);
+
+    if (r == -1) {
+		  error("waitpid: %s", strerror(errno));
+			exit(1);
+    }
+  }
 
   if (WIFEXITED(result) && WEXITSTATUS(result)) {
     error("%s exited with return code %d", progname, WEXITSTATUS(result));
@@ -294,6 +326,26 @@ void run_part(char *progname)
   }
 }
 
+static void handle_signal(int s)
+{
+    /* Do nothing */
+}
+
+/* Catch SIGCHLD with an empty function to interrupt select() */
+static void catch_signals()
+{
+    struct sigaction act;
+    sigset_t set;
+
+    memset(&act, 0, sizeof(act));
+    act.sa_handler = handle_signal;
+    act.sa_flags = SA_NOCLDSTOP;
+    sigaction(SIGCHLD, &act, NULL);
+
+    sigemptyset(&set);
+    sigaddset(&set, SIGCHLD);
+    sigprocmask(SIG_BLOCK, &set, NULL);
+}
 
 /* Find the parts to run & call run_part() */
 void run_parts(char *dirname)
@@ -338,17 +390,17 @@ void run_parts(char *dirname)
     strcat(filename, "/");
     strcat(filename, namelist[i]->d_name);
 
-    result = lstat(filename, &st);
+    result = stat(filename, &st);
     if (result < 0) {
+      error("failed to stat component %s: %s", filename, strerror(errno));
       if (exit_on_error_mode) {
-        error("failed to lstat component %s: %s", filename, strerror(errno));
         exit(1);
       }
       else
-	return;
+        continue;
     }
 
-    if (S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)) {
+    if (S_ISREG(st.st_mode)) {
       if (!access(filename, X_OK)) {
 	if (test_mode) {
 	  printf("%s\n", filename);
@@ -359,7 +411,16 @@ void run_parts(char *dirname)
 	}
 	else {
 	  if (verbose_mode)
-	    fprintf(stderr, "run-parts: executing %s\n", filename);
+	    if (argcount) {
+	      char **a = args;
+
+	      fprintf(stderr, "run-parts: executing %s", filename);
+	      while(*(++a))
+		fprintf(stderr, " %s", *a);
+	      fprintf(stderr, "\n");
+	    } else {
+	      fprintf(stderr, "run-parts: executing %s\n", filename);
+	    }
 	  run_part(filename);
 	  if (exitstatus != 0 && exit_on_error_mode) return;
 	}
@@ -393,6 +454,7 @@ void run_parts(char *dirname)
 /* Process options */
 int main(int argc, char *argv[])
 {
+  custom_ere = NULL;
   umask(022);
   add_argument(0);
 
@@ -452,15 +514,111 @@ int main(int argc, char *argv[])
     error("missing operand");
     fprintf(stderr, "Try `run-parts --help' for more information.\n");
     exit(1);
-  }
-
-  if (list_mode && test_mode) {
+  } else if (list_mode && test_mode) {
     error("--list and --test can not be used together");
     fprintf(stderr, "Try `run-parts --help' for more information.\n");
     exit(1);
+  } else {
+    catch_signals();
+    regex_compile_pattern();
+    run_parts(argv[optind]);
+    regex_clean();
+
+    free(args);
+    free(custom_ere);
+
+    return exitstatus;
   }
+}
 
-  run_parts(argv[optind]);
+/*
+ * Compile patterns used by the application
+ *
+ * In order for a string to be matched by a pattern, this pattern must be
+ * compiled with the regcomp function. If an error occurs, the application
+ * exits and displays the error.
+ */
+static void
+regex_compile_pattern (void)
+{
+    int      err;
+    regex_t *pt_regex;
 
-  return exitstatus;
+    if (regex_mode == RUNPARTS_ERE) {
+
+        if ((err = regcomp(&customre, custom_ere,
+                    REG_EXTENDED | REG_NOSUB)) != 0)
+            pt_regex = &customre;
+
+    } else if (regex_mode == RUNPARTS_LSBSYSINIT) {
+
+        if ( (err = regcomp(&hierre, "^_?([a-z0-9_.]+-)+[a-z0-9]+$",
+                    REG_EXTENDED | REG_NOSUB)) != 0)
+            pt_regex = &hierre;
+
+        else if ( (err = regcomp(&excsre, "^[a-z0-9-].*dpkg-(old|dist|new|tmp)$",
+                    REG_EXTENDED | REG_NOSUB)) != 0)
+            pt_regex = &excsre;
+
+        else if ( (err = regcomp(&tradre, "^[a-z0-9][a-z0-9-]*$", REG_NOSUB))
+                    != 0)
+            pt_regex = &tradre;
+
+    } else if ( (err = regcomp(&classicalre, "^[a-zA-Z0-9_-]+$",
+                    REG_EXTENDED | REG_NOSUB)) != 0)
+        pt_regex = &classicalre;
+
+    if (err != 0) {
+        fprintf(stderr, "Unable to build regexp: %s", \
+                            regex_get_error(err, pt_regex));
+        exit(1);
+    }
+}
+
+/*
+ * Get a regex error.
+ *
+ * This function allocates a buffer to store the regex error description.
+ * If a buffer cannot be allocated, then the use of xmalloc will end the
+ * program.
+ *
+ * @errcode: return error code from a one of the regex functions
+ * @compiled: compile pattern which causes the failure
+ *
+ * It returns a pointer on the current regex error description.
+ */
+static char *
+regex_get_error(int errcode, regex_t *compiled)
+{
+    size_t  length;
+    char     *buf;
+
+    length = regerror(errcode, compiled, NULL, 0);
+    buf    = malloc(length);
+    if (buf == 0) {
+        error("Virtual memory exhausted\n");
+        exit(1);
+    }
+
+    regerror(errcode, compiled, buf, length);
+
+    return buf;
+}
+
+/*
+ * Clean the compiled patterns according to the current regex_mode
+ */
+static void
+regex_clean(void)
+{
+    if (regex_mode == RUNPARTS_ERE)
+        regfree(&customre);
+
+    else if (regex_mode == RUNPARTS_LSBSYSINIT) {
+        regfree(&hierre);
+        regfree(&excsre);
+        regfree(&tradre);
+
+    } else
+        regfree(&classicalre);
 }
